# SpeedBurst by David Bowland
# ./addons/eventscripts/speedburst/speedburst.py

# >>> To configure this addon please see speedburst.cfg (created when the addon is first loaded) <<<

"""
Allows players to have a sudden, triggered speed burst and a chargeable speed burst.
Requires only ES 2.0.0.250c+
"""


from __future__ import division

import cfglib
import cmdlib
import es
import gamethread
import langlib
import os.path
import playerlib
import time

import psyco
psyco.full()


info = es.AddonInfo()
info.name     = 'SpeedBurst'
info.version  = '1'
info.basename = 'speedburst'
info.url      = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author   = 'SuperDave'

###

config = cfglib.AddonCFG(os.path.join(es.getAddonPath(info.basename), info.basename + '.cfg'))

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename': info.basename})
config.text(info.url)
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('\n')
config.text('To configure language settings for %s please see %s_languages.ini' % (info.name, info.basename))
config.text('\n')

config.text('***** Speed burst options *****')
cvar_burst_on         = config.cvar('speedburst_burst_on',         1,        '0 = no short speed burst, 1 = speed burst possible with say command')
cvar_burst_trigger    = config.cvar('speedburst_burst_trigger',    '!speed', 'Say command to trigger a speed burst')
cvar_burst_multiplier = config.cvar('speedburst_burst_multiplier', 3,        'Velocity multiplier for speed burst, 1 = normal speed')
cvar_burst_timelimit  = config.cvar('speedburst_burst_timelimit',  15,       'Number of seconds players must wait between using the speed burst')
config.text('\n')

config.text('***** Speed charge options *****')
cvar_charge_on      = config.cvar('speedburst_charge_on',     1,            '0 = no chargeable burst, 1 = allow chargeable speed burst')
cvar_charge_trigger = config.cvar('speedburst_chargetrigger', 'speedburst', 'Client command to charge speed bursts')
config.text('Bind a key to the charge trigger with a + in front of it. Hold down the key to charge the speed')
config.text('burst and release the key to use the charged speed.')
cvar_charge_chargetime     = config.cvar('speedburst_charge_chargetime',     15,  'Number of seconds to completely charge a speed burst')
cvar_charge_progressupdate = config.cvar('speedburst_charge_progressupdate', 1,   'Number of seconds between updating the charge hudhint -- must be greater than zero and less than or equal to one')
cvar_charge_minmultiplier  = config.cvar('speedburst_charge_minmultiplier',  0.7, 'Multiplier for a completely uncharged speed burst')
cvar_charge_maxmultiplier  = config.cvar('speedburst_charge_maxmultiplier',  6,   'Multiplier for a completely charged speed burst')
config.text('\n')

config.text('***** Announcement options *****')
cvar_announce_command        = config.cvar('speedburst_announce_command',        2, 'Announcing speed commands: 0 = no announcement, 1 = in console, 2 = in chat area')
cvar_announce_toosoon_chat   = config.cvar('speedburst_announce_toosoon_chat',   2, 'Announcing using speed burst too soon after last: 0 = no change, 1 = in console, 2 = in chat area')
cvar_announce_toosoon_center = config.cvar('speedburst_announce_toosoon_center', 1, 'Announcing using speed burst too soon after last: 0 = no change, 1 = in center text')
cvar_announce_toosoon_hud    = config.cvar('speedburst_announce_toosoon_hud',    1, 'Announcing using speed burst too soon after last: 0 = no change, 1 = as a hud hint')
config.text('\n')

config.text('***** Notes *****')
config.text('-A maximum velocity of 2048 is imposed due to Source restrictions.')
config.text('-Due to the fact speed bursts are multipliers the faster a player is moving when a burst is used the bigger the net speed gain.')
config.text(' Use a speed burst after jumping to reach extraordinary heights!')
config.text('-Because of the jumping possibilities with speed bursts, this addon is fun to couple with Jump Stats:')
config.text(' http://addons.eventscripts.com/addons/view/jumpstats')
config.text('\n')

config.text('***** Event documentation *****')
config.text('')
config.text('The following events will fire while SpeedBurst is loaded:')
config.text('')
config.text('speedburst_burst')
config.text('- Fires when a player uses a speed burst.')
config.text('   event_var(userid)          - Player who used the speed burst')
config.text('')
config.text('speedburst_startcharge')
config.text('   event_var(userid)          - Player who began charging his or her speed burst')
config.text('')
config.text('speedburst_endcharge')
config.text('   event_var(userid)          - Player who stopped charging his or her speed burst')
config.text('   event_var(time)            - Number of seconds the player charged the speed burst')
config.text('   event_var(multiplier)      - Speed multiplier the player received')
config.text('')
config.text('Each event also provides event_var(es_steamid), event_var(es_username), and all the other event_vars that accompany')
config.text('event_var(userid). More information can be found here: http://www.eventscripts.com/pages/Extended_event_variables')

config.write()

###

ini = cfglib.AddonINI(os.path.join(es.getAddonPath(info.basename), info.basename + '_languages.ini'))

ini.setInitialComments(['# ./addons/eventscripts/%(basename)s/%(basename)s_languages.ini' % {'basename': info.basename},
 '# >>> For general addon configuration see ' + info.basename + '.cfg <<<', '', '# DO NOT translate words with $ or # in front!',
 '# For a list of language abbreviations please see ./addons/eventscripts/_libs/python/deflangs.ini'])

ini.addGroup('burst too soon')
ini.addValueToGroup('burst too soon', 'en', '#defaultYou must wait #lightgreen$seconds seconds#default between using speed bursts.')

ini.addGroup('burst command')
ini.addValueToGroup('burst command', 'en', '#defaultSay  #lightgreen$trigger#default  to receive a speed burst.')

ini.addGroup('charge command')
ini.addValueToGroup('charge command', 'en', '#defaultBind a key to  #lightgreen+$trigger#default  to charge a speed burst.')

ini.addGroup('charge in progress')
ini.addValueToGroup('charge in progress', 'en', 'Speed charge: $percent%')

ini.write()
lang_text = langlib.Strings(ini)

###

class SpeedBase(object):
   """ Base class providing speed and player control functionality """
   def __init__(self):
      self.players = {}

   """ Player functions """

   def clearPlayers(self):
      """ Clears the dictionary of players """
      self.players.clear()

   def removePlayer(self, userid):
      """ Removes a player from the dictionary of players """
      userid = int(userid)
      if userid in self.players:
         del self.players[userid]

   def playerSpawn(self, userid):
      """ Calls removePlayer so spawning players are removed from the dictionary of players """
      self.removePlayer(userid)

   """ Static methods """

   @staticmethod
   def commandAnnounce(userid, text):
      """ Announces text to the chat and console """
      if cvar_announce_command == 2:
         es.tell(userid, '#multi', text)
      elif cvar_announce_command == 1:
         es.cexec(userid, 'echo ' + text)

   @staticmethod
   def multiplyVelocity(userid, multiplier):
      """ Sets the userid's speed multiplier """
      es.setplayerprop(userid, 'CBasePlayer.localdata.m_vecBaseVelocity', ','.join(
       [str(min(2048, float(es.getplayerprop(userid, 'CBasePlayer.localdata.m_vecVelocity[%s]' % x))) *
       (multiplier if x != 2 else multiplier / 2)) for x in xrange(3)]))

   @staticmethod
   def hudHint(userid, message):
      """ Sends a player a HUD hint """
      es.usermsg('create', 'speedburst_hud', 'HintText')
      es.usermsg('write', 'byte', 'speedburst_hud', -1)
      es.usermsg('write', 'string', 'speedburst_hud', message)
      es.usermsg('send', 'speedburst_hud', userid)
      es.usermsg('delete', 'speedburst_hud')

   @staticmethod
   def removeTags(text):
      """ Removes #lightgreen, #green and #default tags from text """
      return text.replace('#lightgreen', '').replace('#green', '').replace('#default', '')

###

class SpeedBurst(SpeedBase):
   saycommand = None

   def registerSayCommand(self, command):
      """ Registers the burst client command """
      self.unregisterSayCommand()
      cmdlib.registerSayCommand(command, self.triggerBurst, 'Say command to trigger a short speed burst')
      self.saycommand = command

   def unregisterSayCommand(self):
      """ Unregisters the burst client command """
      if self.saycommand:
         cmdlib.unregisterSayCommand(self.saycommand)
         self.saycommand = None

   def triggerBurst(self, userid, args):
      """ Gives the calling player a speed burst """
      player = playerlib.getPlayer(userid)
      if player.getAttribute('isdead'): # Must be alive
         return
      if userid not in self.players or self.players[userid] < time.time() - cvar_burst_timelimit:
         self.multiplyVelocity(userid, cvar_burst_multiplier)
         self.players[userid] = time.time()

         # Fire speed burst event
         es.event('initialize', 'speedburst_burst')
         es.event('setint',     'speedburst_burst', 'userid', userid)
         es.event('fire',       'speedburst_burst')

      # The player is trying to use the command too quickly
      else:
         text = lang_text('burst too soon', {'seconds': cvar_burst_timelimit}, player.getLanguage())
         if cvar_announce_toosoon_chat == 1:
            es.cexec(userid, 'echo ' + self.removeTags(text))
         elif cvar_announce_toosoon_chat:
            es.tell(userid, '#multi', text)

         text = self.removeTags(text)
         if cvar_announce_toosoon_center:
            es.centertell(userid, text)
         if cvar_announce_toosoon_hud:
            self.hudHint(userid, text)

   def playerSpawn(self, userid):
      """ Announce the say command to trigger speed burst """
      if self.saycommand:
         player = playerlib.getPlayer(userid)
         if not player.getAttribute('isdead'):
            self.commandAnnounce(userid, lang_text('burst command', {'trigger': self.saycommand}, player.getLanguage()))
      super(SpeedBurst, self).playerSpawn(userid)

burst = SpeedBurst()

###

class SpeedCharge(SpeedBase):
   clientcommand = None

   def clearPlayers(self):
      """ Cancels delays for players charging the speed burst """
      for userid in self.players:
         gamethread.cancelDelayed(self.getDelayName(userid))
      super(SpeedCharge, self).clearPlayers()

   def removePlayer(self, userid):
      """ Removes a player from the speed charge dictionary """
      gamethread.cancelDelayed(self.getDelayName(userid))
      super(SpeedCharge, self).removePlayer(userid)

   def registerClientCommands(self, command):
      """ Registers the charge client commands """
      self.unregisterClientCommands()
      cmdlib.registerClientCommand('+' + command, self.startCharge, 'Client command to start speed charge')
      cmdlib.registerClientCommand('-' + command, self.endCharge,   'Client command to end speed charge')
      self.clientcommand = command

   def unregisterClientCommands(self):
      """ Unregisters the charge client commands """
      if self.clientcommand:
         cmdlib.unregisterClientCommand('+' + self.clientcommand)
         cmdlib.unregisterClientCommand('-' + self.clientcommand)
         self.clientcommand = None

   def startCharge(self, userid, args):
      """ Starts the speed charge """
      if playerlib.getPlayer(userid).isdead: # Must be alive
         return
      self.players[userid] = time.time() # We need to know what time the player started charging
      self.chargeStep(userid)

      # Fire start charge event
      es.event('initialize', 'speedburst_startcharge')
      es.event('setint',     'speedburst_startcharge', 'userid', userid)
      es.event('fire',       'speedburst_startcharge')

   def endCharge(self, userid, args):
      """ Ends the speed charge """
      if userid in self.players:
         charge_time = time.time() - self.players[userid]
         charge_percent = min(1, charge_time / cvar_charge_chargetime)
         multiplier = (float(cvar_charge_maxmultiplier) - cvar_charge_minmultiplier) * charge_percent + cvar_charge_minmultiplier

         self.hudHint(userid, lang_text('charge in progress', {'percent': int(charge_percent * 100)}, playerlib.getPlayer(userid).getLanguage()))
         self.multiplyVelocity(userid, multiplier)
         del self.players[userid]

         # Fire end charge event
         es.event('initialize', 'speedburst_endcharge')
         es.event('setint',     'speedburst_endcharge', 'userid',     userid)
         es.event('setfloat',   'speedburst_endcharge', 'time',       charge_time)
         es.event('setfloat',   'speedburst_endcharge', 'multiplier', multiplier)
         es.event('fire',       'speedburst_endcharge')

      gamethread.cancelDelayed(self.getDelayName(userid))

   def chargeStep(self, userid):
      """ Moves the status bar and displays the charging HUD hint """
      if userid in self.players:
         charge_percent = min(1, (time.time() - self.players[userid]) / cvar_charge_chargetime) * 100
         progress_bar   = '[]' * int(charge_percent // 10)
         if charge_percent % 10 >= 5:
            progress_bar += '['

         header = lang_text('charge in progress', {'percent': int(charge_percent)}, playerlib.getPlayer(userid).getLanguage()).center(len(progress_bar))
         progress_bar = progress_bar.center(len(header))
         self.hudHint(userid, header + '\n' + progress_bar)
         gamethread.delayedname(max(0.01, min(1, float(cvar_charge_progressupdate))), self.getDelayName(userid), self.chargeStep, userid)

   def playerSpawn(self, userid):
      """ Displays the charge client commands """
      if self.clientcommand:
         player = playerlib.getPlayer(userid)
         if not player.getAttribute('isdead'):
            self.commandAnnounce(userid, lang_text('charge command', {'trigger': self.clientcommand}, player.getLanguage()))
      super(SpeedCharge, self).playerSpawn(userid)

   @staticmethod
   def getDelayName(userid):
      return 'speedburst_' + str(userid)

charge = SpeedCharge()

###

def _loadEvents():
   """ Loads the .res file for the events """
   if os.path.isfile(os.path.join(es.getAddonPath(info.basename), info.basename + '.res')):
      es.loadevents('declare', 'addons/eventscripts/%(basename)s/%(basename)s.res' % {'basename': info.basename})

   else:
      events = {'speedburst_burst': {'userid': 'short'}, 'speedburst_startcharge': {'userid': 'short'},
       'speedburst_endcharge': {'userid': 'short', 'time': 'float', 'multiplier': 'float'}}

      if not os.path.isfile(os.path.join(es.getAddonPath(info.basename), + 'es_' + info.basename + '_events_db.txt')):
         es.keygroupcreate(info.basename + '_events')
         for event_name in events:
            es.keycreate(info.basename + '_events', event_name)
            for event_var in events[event_name]:
               es.keysetvalue(info.basename + '_events', event_name, event_var, events[event_name][event_var])
         es.keygroupsave(info.basename + '_events', '|' + info.basename)
         es.keygroupdelete(info.basename + '_events')
      es.loadevents('declare', 'addons/eventscripts/' + info.basename + '/es_' + info.basename + '_events_db.txt')
      es.dbgmsg(0, info.name + ': Unable to load ' + info.basename + '.res! Please ensure it is in the ./' + info.basename + '/ directory.')

###

def load():
   config.execute()
   _loadEvents()

   if cvar_burst_on:
      burst.registerSayCommand(str(cvar_burst_trigger))
   if cvar_charge_on:
      charge.registerClientCommands(str(cvar_charge_trigger))


def es_map_start(event_var):
   _loadEvents()
   burst.clearPlayers()
   charge.clearPlayers()


def round_start(event_var):
   burst.clearPlayers()
   charge.clearPlayers()


def player_spawn(event_var):
   userid = int(event_var['userid'])
   burst.playerSpawn(userid)
   charge.playerSpawn(userid)


def player_death(event_var):
   userid = int(event_var['userid'])
   burst.removePlayer(userid)
   charge.removePlayer(userid)


def player_disconnect(event_var):
   userid = int(event_var['userid'])
   burst.removePlayer(userid)
   charge.removePlayer(userid)


def unload():
   burst.unregisterSayCommand()
   charge.unregisterClientCommands()

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)